/*******************************************************************************
  Copyright(c) 2000 - 2003 Radu Corlan. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information: radu@corlan.net
*******************************************************************************/

// ccd_frame.c: frame operation functions
// $Revision: 1.4 $
// $Date: 2003/05/19 01:09:37 $

#include <math.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>

#include "ccd.h"
//#include "cpxcon_regs.h"

#define MAX_FLAT_GAIN 1.5	// max gain accepted when flatfielding
#define FITS_HROWS 36	// number of fits header lines in a block

// new frame creates a frame of the specified size
struct ccd_frame *new_frame(unsigned size_x, unsigned size_y)
{
	return new_frame_fr(NULL, size_x, size_y);
}

// clone_frame produces a newly allocced copy of a frame
struct ccd_frame *clone_frame(struct ccd_frame *fr)
{
	struct ccd_frame *new;
	int all, i;

	new = new_frame_fr(fr, 0, 0);
	if (new == NULL)
		return NULL;
// now copy the data
//	d3_printf("nffr\n");
	all = new->geom.sz_x * new->geom.sz_y;
	memcpy(new->dat, fr->dat, all * fr->geom.pix_size);
	return new;
}

// new_frame_fr creates and allocates a frame copying it's geometry from
// a given frame; when non-zero, the size params overwrite the default.
// returns a pointer to the new image header, or NULL if the request falls
// magic is copied from the original
struct ccd_frame *new_frame_fr(struct ccd_frame* fr, unsigned size_x, unsigned size_y)
{
	struct ccd_frame *new;
	void *data;

	new = new_frame_head_fr(fr, size_x, size_y);
	if (new == NULL)
		return NULL;
//	d3_printf("nfrhfr\n");

	data = malloc(new->geom.sz_x * 
		      new->geom.sz_y * 
		      new->geom.pix_size);

	if (data == NULL) {
		free(new);
		return NULL;
	}
	new->dat = data;
	return new;
}

// new_frame_head_fr is the same as new_frame_fr, only it does not allocate space for the
// data array, only for the header
// if the fr parameter is NULL, the function creates a frame header with only
// the magic and geometry fields set. It assumes 16bits/pix.
struct ccd_frame *new_frame_head_fr(struct ccd_frame* fr, unsigned size_x, unsigned size_y)
{
	struct ccd_frame *hd;
	void *var;

	hd = calloc(1, sizeof(struct ccd_frame) );
	if (hd == NULL)
		return NULL;

	if (fr != NULL) {
		memcpy(hd, fr, sizeof(struct ccd_frame) );
	} 
	hd->geom.pix_size = 4;

	if (hd->nvar) { //we need to alloc space for variables
		var = calloc(hd->nvar * sizeof(FITS_row), 1);
		if (var == NULL) {
			free(hd);
			return NULL;
		}
		memcpy(var, hd->var, hd->nvar * sizeof(FITS_row));
		hd->var = var;
	} else {
		hd->var = NULL;
		hd->nvar = 0;
	}

	hd->stats.hist.hdat = NULL;
	hd->stats.statsok = 0;

	hd->magic = UNDEF_FRAME;
	if (size_x != 0)
		hd->geom.sz_x = size_x;
	if (size_y != 0)
		hd->geom.sz_y = size_y;
	if (!hd->fim.wcsset) {
		hd->fim.xref = 0.0;
		hd->fim.yref = 0.0;
		hd->fim.xrefpix = 0.0;
		hd->fim.yrefpix = 0.0;
		hd->fim.xinc = -0.000263;
		hd->fim.yinc = -0.000263;
		hd->fim.rot = 0.0;
	}
	if (!hd->fim.bsset) {
		hd->fim.bscale = 1.0;
		hd->fim.bzero = 0.0;
	}
	if (fr == NULL) {
		hd->exp.scale = 1.0;
		hd->exp.bias = 0.0;
		hd->exp.rdnoise = 1.0;
		hd->exp.flat_noise = 0.0;
	}
//	d3_printf("new_head_fr returns\n");
	return hd;
}

// free_frame frees a frame completely (data array and header)
void free_frame(struct ccd_frame *fr)
{
	if (fr) {
		free(fr->var);
		free(fr->stats.hist.hdat);
		free(fr->dat);
		free(fr);
	}
}

// free_frame_data frees the data array of a frame
void free_frame_data(struct ccd_frame *fr)
{
	free(fr->dat);
	free(fr->stats.hist.hdat);
	fr->stats.hist.hdat = NULL;
	fr->dat = NULL;
	fr->data_valid = 0;
}

// alloc_frame_data allocated the data area for a frame according to the 
// parameters in the supplied header; returns ERR_ALLOC if it cannot allocate the space, 0 for ok.
int alloc_frame_data(struct ccd_frame *fr)
{
	void *data;

	data = malloc(fr->geom.sz_x * 
		      fr->geom.sz_y * 
		      fr->geom.pix_size);

	if (data == NULL) {
		return ERR_ALLOC;
	} else {
		fr->dat = data;
		return 0;
	}


}

// frame_stats fills the statistics buffer for the frame
#define H_MIN 0.0 // leftmost bin value
#define HIST_OFFSET 1000.0 // offset to accomodate small negative values in histogram
#define H_MAX 65535.0 // rightmost bin value

int frame_stats(struct ccd_frame *hd)
{
	double sum, sumsq, min, max, avg, sigma;
	int i, is, ie, j, k, ef;
	unsigned all, s, e;
	float *data;
	float v;
	unsigned hsize;
	double hmin, hstep;
	unsigned binmax;
	unsigned *hdata;
	int b, n;
	double cs, ce;
	double median, c, bv;

	if (hd->stats.hist.hdat == NULL) { // we need to allocate the histogram data
		if (hd->stats.hist.hsize == 0)
			hd->stats.hist.hsize = 65536;
		hd->stats.hist.hdat = (unsigned *)calloc(hd->stats.hist.hsize, sizeof(unsigned));
		if (hd->stats.hist.hdat == NULL) {
			err_printf("frame_stats: histogram alloc failed\n");
			return ERR_ALLOC;
		}
	}

	hdata = hd->stats.hist.hdat;

	hsize = hd->stats.hist.hsize;
	hmin = H_MIN;
	hstep = (H_MAX - H_MIN) / hsize;
	binmax = 0;

	sum = 0.0;
	sumsq = 0.0;

// do the reading and calculate stats
	all = hd->geom.sz_x * (hd->geom.sz_y); 
	data = (float *)hd->dat;
	min = *data;
	max = *data;
// TODO remove this skip
//	data += 4 * STATS_SKIP * hd->geom.sz_x;
	for (i = 0; i < all; i++) {
		v = *data++;
		if (isnan(v))
			d3_printf("found NaN at %d", i);
//		if (v < -HIST_OFFSET)
//			v = -HIST_OFFSET;
//		if (v > H_MAX)
//			v = H_MAX;
		b = HIST_OFFSET + floor( (v - hmin) / hstep );
		if (b < 0)
			b = 0;
		if (b >= hsize)
			b = hsize - 1;
		hdata[b] ++;
		if (hdata[b] > binmax)
			binmax = hdata[b];
		if (v > max) {
			max = v;
//			d3_printf("max = %.1f", max);
		}
		else if (v < min) {
			min = v;
//			d3_printf("min = %.1f", min);
		}
		sum += v;
		sumsq += v * v;
	}

	hd->data_valid = 1;
//	hd->magic = UNDEF_FRAME;
	hd->stats.min = min;
	hd->stats.max = max;
	hd->stats.avg = sum / (1.0 * all);
	hd->stats.sigma = sqrt(sumsq / (1.0 * all) - hd->stats.avg * hd->stats.avg);
	hd->stats.hist.binmax = binmax;

//	d3_printf("all:%d sum:%.1f\n", all, sum);

// scan the histogram to get the median, cavg and csigma

	sum = 0.0;
	sumsq = 0.0;
	b=0;
	n=0;

	bv = hmin - HIST_OFFSET;
	s = all / POP_CENTER;
	e = all - all / POP_CENTER;
	c = all / 2;
	for (i = 0; i < hsize; i++) {
		b += hdata[i];
		bv += hstep;
		if (b < s) {
			is = i;
			continue;
		}
		if (b > e) {
			ie = i;
			break;
		}
		if (b - hdata[i] < c && b >= c) { // we are at the median point
			if (hdata[i] != 0) { // get the interpolated median
				median = (b - c) / hdata[i] * hstep + i + hmin - HIST_OFFSET;
//				info_printf("M: %.3f\n", median);
			} else {
				median = (i - 0.5) * hstep + hmin - HIST_OFFSET;
			}
		}
		n += hdata[i];
		sum += hdata[i] * bv;
		sumsq += hdata[i] * bv * bv;
	}
	i = is;
	hd->stats.hist.cst = i;
	cs = hmin + hstep * i - HIST_OFFSET;

//	info_printf("new cavg %.3f csigm %.3f median %.3f, sigma, %.3f ", sum / n, 
//		    sqrt(sumsq / n - sqr(sum / n)), median, hd->stats.sigma);

	hd->stats.cavg = sum / n;
	hd->stats.csigma = 2 * sqrt(sumsq / n - sqr(sum / n));
	hd->stats.median = median;

	hd->stats.statsok = 1;
//	info_printf("min=%.2f max=%.2f avg=%.2f sigma=%.2f cavg=%.2f csigma=%.2f\n", 
//		    hd->stats.min, hd->stats.max, hd->stats.avg, hd->stats.sigma, 
//		    hd->stats.cavg, hd->stats.csigma);
	return 0;
}

// check if a filename has a valid fits extension, or append one if it doesn't
// do not exceed fnl string length
// return 1 if the name was changed

static int fits_filename(char *fn, int fnl)
{
	int len;
	len = strlen(fn);
	if (len + 5 >= fnl)
		return 0;
	if (strcmp(fn + len - 5, ".fits") == 0) 
		return 0;
	if (strcmp(fn + len - 4, ".fit") == 0) 
		return 0;
	if (strcmp(fn + len - 4, ".fts") == 0) 
		return 0;
	if (strcmp(fn + len - 5, ".FITS") == 0) 
		return 0;
	if (strcmp(fn + len - 4, ".FTS") == 0) 
		return 0;
	if (strcmp(fn + len - 4, ".FIT") == 0) 
		return 0;
	strcat(fn, ".fits");
	return 1;
}

#define MAX_FILENAME 1024

// read_fits_file reads a fits file from disk and creates a new frame 
// holding the data from the file. 
struct ccd_frame *read_fits_file(char *filename)
{
	char lb[FITS_HCOLS + 1];
	char fn[MAX_FILENAME + 1];
	FILE *fp;
	short v[FITS_HCOLS * FITS_HROWS / 2];
	short ds;
	int i, j, k, ef;
	unsigned all;
	unsigned naxis;
	struct ccd_frame* hd;
	float *data;
	int bitpix;
	FITS_row *cp;

	double bz, bs;
	float d;

	ef = 0;
	lb[80] = 0;

//	d3_printf("A");



// pad the filename and open the file
	strncpy(fn, filename, MAX_FILENAME);

	fits_filename(fn, MAX_FILENAME);

	fp = fopen(fn, "r");
	if (fp == NULL) {
		err_printf("\nCannot open file: %s\n", fn);
		return NULL;
	}

//now allocate the header for the new frame
	hd = new_frame_head_fr(NULL, 0, 0);
		if (hd == NULL) {
			err_printf("read_fits_file: cannot allocate header for frame %s\n", fn);
			return NULL;
		}
	for (i = 0; i < 6; i++)	// at most 6 header blocks
	{
		for (j = 0; j < FITS_HROWS; j++)	// 36 cards per block
		{
			for (k = 0; k < FITS_HCOLS; k++)	// chars per card
				lb[k] = fgetc(fp);
// now parse the fits header lines
			printf("%s\n", lb);
			if (strncmp(lb, "END", 3) == 0) 
				ef = 1;
			else if (sscanf(lb, "NAXIS   = %d", &naxis) ) 
				;
			else if (sscanf(lb, "NAXIS1  = %d", &hd->geom.sz_x) ) 
				;
			else if (sscanf(lb, "NAXIS2  = %d", &hd->geom.sz_y) ) 
				;
			else if (sscanf(lb, "BITPIX  = %d", &hd->geom.pix_size) ) 
				;
 			else if (sscanf(lb, "CRPIX1  = %f", &d) ) {
				hd->fim.wcsset = 1;
				hd->fim.xrefpix = d;
			}
			else if (sscanf(lb, "CRVAL1 = %f", &d) ) {
				hd->fim.wcsset = 1;
				hd->fim.xref = d;
			}
			else if (sscanf(lb, "CDELT1 = %f", &d) ) {
				hd->fim.wcsset = 1;
				hd->fim.xinc = d;
			}
			else if (sscanf(lb, "CROTA1 = %f", &d) ) {
				hd->fim.wcsset = 1;
				hd->fim.rot = d;
			}
 			else if (sscanf(lb, "CRPIX2 = %f", &d) ) {
				hd->fim.wcsset = 1;
				hd->fim.yrefpix = d;
			}
			else if (sscanf(lb, "CRVAL2 = %f", &d) ) {
				hd->fim.wcsset = 1;
				hd->fim.yref = d;
			}
			else if (sscanf(lb, "CDELT2 = %f", &d) ) {
				hd->fim.wcsset = 1;
				hd->fim.yinc = d;
			}
			else if (sscanf(lb, "CROTA2 = %f", &d) ) {
				hd->fim.wcsset = 1;
				hd->fim.rot = d;
			}
			else if (sscanf(lb, "EQUINOX = %d", &hd->fim.equinox) ) 
				;
			else if (sscanf(lb, "BSCALE = %f", &d) ) {
				hd->fim.bsset = 1;
				hd->fim.bscale = d;
			}
			else if (sscanf(lb, "BZERO = %f", &d) ) {
				hd->fim.bsset = 1;
				hd->fim.bzero = d;
			}
			else if (!strncmp(lb, "CTYPE1", 6) ) 
				;
			else if (!strncmp(lb, "CTYPE2", 6) ) 
				;
			else if (!strncmp(lb, "SIMPLE", 6) ) 
				;
			else if (!ef) { //add the line to the unprocessed list
				cp = realloc(hd->var, (hd->nvar + 1) * FITS_HCOLS);
				if (cp == NULL) {
					err_printf("cannot alloc space for fits header line, skipping\n");
				} else {
					hd->var = cp;
					memcpy(hd->nvar + cp, lb, FITS_HCOLS);
					hd->nvar ++;
				}
			}
		}
		if (ef == 1)
			break;
	}

	if (ef == 0) {
		err_printf("%s : Bad FITS format; cannot find END keyword\n", fn);
		goto err_exit;
	}

// checking header data
	if (naxis != 2) {
		err_printf("%s : bad NAXIS = %d (!= 2)\n", fn, naxis);
		goto err_exit;
	}
	if (hd->geom.pix_size == 16) {
		hd->geom.pix_size = 4;
	} else {
		err_printf("%s : bad BITPIX = %d (must be 16)\n", fn, hd->geom.pix_size);
		goto err_exit;
	}
	if (hd->geom.sz_x <= 0) {
		err_printf("%s : bad NAXIS1 = %d \n", fn, hd->geom.sz_x);
		goto err_exit;
	}
	if (hd->geom.sz_y <= 0) {
		err_printf("%s : bad NAXIS2 = %d \n", fn, hd->geom.sz_y);
		goto err_exit;
	}

	if (alloc_frame_data(hd)) {
		err_printf("read_fits_file: cannot allocate data for frame %s\n", fn);
		goto err_exit;
	}

	//	info_printf("Reading FITS file: %s %d x %d x %d\n", 
	//		    fn, hd->geom.sz_x, hd->geom.sz_y, 8*hd->geom.pix_size);

// check for any required scaling/shifting
	if (hd->fim.bsset) {
		if (hd->fim.bscale == 0.0) {
			err_printf("bad BSCALE (0) - using 1.0\n");
			hd->fim.bscale = 1.0;
		}
		bz = hd->fim.bzero;
		bs = hd->fim.bscale;
	} else {
		bz = 0;
		bs = 1.0;
	}

// do the reading and calculate stats
	all = hd->geom.sz_x * hd->geom.sz_y; 
	data = (float *)(hd->dat);
	j=0;

	while(j < all) {
		k = fread (v, 1, FITS_HCOLS * FITS_HROWS, fp);
		if (k < 0) {
			err_printf("data is short!\n");
			break;
		} else 
		for(i = 0; i < FITS_HCOLS * FITS_HROWS / 2 && j < all; i++) {
			ds = (((v[i] >> 8) & 0x0ff) | ((v[i] << 8) & 0xff00));
			*data++ = bs * ds + bz;
			j++;
		}
	}

//	d3_printf("B");


	hd->data_valid = 1;
	hd->magic = UNDEF_FRAME;

	frame_stats(hd);

//	info_printf("min=%.2f max=%.2f avg=%.2f sigma=%.2f\n", 
//		    hd->stats.min, hd->stats.max, hd->stats.avg, hd->stats.sigma);

	strncpy(hd->name, fn, 255);
//	libip_head(hd);

// 	now read the skip/bin/exp/temperature info from the headers
	cp = fits_keyword_lookup(hd, "CCDBIN1");
	if (cp != NULL)
		if (!sscanf((char *)cp, "%*s = %d", &(hd->geom.bin_x)))
			hd->geom.bin_x = 1;
	cp = fits_keyword_lookup(hd, "CCDBIN2");
	if (cp != NULL)
		if (!sscanf((char *)cp, "%*s = %d", &(hd->geom.bin_y)))
			hd->geom.bin_y = 1;
	cp = fits_keyword_lookup(hd, "CCDSKIP1");
	if (cp != NULL)
		if (!sscanf((char *)cp, "CCDSKIP1= %d", &(hd->geom.x_skip)))
			hd->geom.x_skip = 0;
	cp = fits_keyword_lookup(hd, "CCDSKIP2");
	if (cp != NULL)
		if (!sscanf((char *)cp, "CCDSKIP2= %d", &(hd->geom.y_skip)))
			hd->geom.y_skip = 0;
	cp = fits_keyword_lookup(hd, "EXPTIME");
	if (cp != NULL)
		if (sscanf((char *)cp, "%*s = %f", &d) )
			hd->exp.exp_time = (int)floor(1000.0 * d);
	cp = fits_keyword_lookup(hd, "SNS_TEMP");
	if (cp != NULL) {
		if (sscanf((char *)cp, "SNS_TEMP= %f", &d) )
			hd->exp.temperature = (int)floor(10.0 * (d - 273.0));
	}

// read in the 'noise' parameters
	cp = fits_keyword_lookup(hd, "ELADU");
	if (cp != NULL) {
		if (sscanf((char *)cp, "%*s = %f", &d) )
			hd->exp.scale = (d);
	} else { // we assume 2e/adu
		hd->exp.scale = 2.0;
		fits_add_history(hd, "MISSING ELADU, ASSUMING 2 ELECTRONS/ADU");
	}
	cp = fits_keyword_lookup(hd, "RDNOISE");
	if (cp != NULL) {
		if (sscanf((char *)cp, "%*s = %f", &d) )
			hd->exp.rdnoise = (d);
//		d3_printf("load_fits_file: got rd_noise= %.3f\n", hd->exp.rdnoise);
	} else { // estimate noise from the frame
		hd->exp.rdnoise = hd->stats.csigma;
		fits_add_history(hd, "MISSING RDNOISE, USING CSIGMA VALUE INSTEAD");
	}
	cp = fits_keyword_lookup(hd, "FLNOISE");
	if (cp != NULL) {
		if (sscanf((char *)cp, "%*s = %f", &d) )
			hd->exp.flat_noise = (d);
	} else { // estimate noise from the frame
		hd->exp.flat_noise = 0;
		fits_add_history(hd, "MISSING FLNOISE, ASSUMING 0");
	}
	cp = fits_keyword_lookup(hd, "DCBIAS");
	if (cp != NULL) {
		if (sscanf((char *)cp, "%*s = %f", &d) )
			hd->exp.bias = (d);
	} else { // estimate noise from the frame
		hd->exp.bias = 0;
		fits_add_history(hd, "MISSING DCBIAS, ASSUMING 0");
	}



	fclose(fp);
//	d3_printf("C");
	return hd;

err_exit: 
	free_frame(hd);
	return NULL;
}


// write a frame to disk as a fits file 

int write_fits_frame(struct ccd_frame *fr, char *filename)
{
	char lb[MAX_FILENAME];
	FILE *fp;
	int i, j, k, v;
	unsigned all;
	struct tm *t;
	float *dp;
	int jdi;
	double jd;

	unsigned u = 0;

	lb[80] = 0;

	strncpy(lb, filename, MAX_FILENAME);

	fits_filename(lb, MAX_FILENAME);

	fp = fopen(lb, "w");
	if (fp == NULL) {
		err_printf("\nwrite_fits_frame: Cannot open file: %s for writing\n",
			filename);
		return (ERR_FILE);
	}

	t = gmtime(&(fr->exp.exp_start.tv_sec));
	sprintf(lb, "'%d-%02d-%02dT%02d:%02d:%02d.%02d'", 1900 + t->tm_year, t->tm_mon + 1, 
		t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec, 
		fr->exp.exp_start.tv_usec / 10000);
// do the julian date (per hsaa p107)
	i = 1900 + t->tm_year;
	j = t->tm_mon + 1;
	k = t->tm_mday;

	jdi = k - 32075 + 1461 * (i + 4800 + (j - 14) / 12) / 4
		+ 367 * (j - 2 - (j - 14) / 12 * 12) / 12
		- 3 * (( i + 4900 + (j - 14) / 12) / 100) / 4;

	jd = jdi - 0.5 + t->tm_hour / 24.0 + t->tm_min / (24.0 * 60.0) 
		+ t->tm_sec / (24.0 * 3600.0) + fr->exp.exp_start.tv_usec / 
		(1000000.0 * 24.0 * 3600.0);


	if (fr->stats.statsok == 0)
		frame_stats(fr);

	i = 0;
	i++; fprintf(fp, "%-8s= %20s / %-40s       ", "SIMPLE", "T",
		"Standard FITS format");
	i++; fprintf(fp, "%-8s= %20d / %-40s       ", "BITPIX", 16,
		"Bits per pixel");
	i++; fprintf(fp, "%-8s= %20d   %-40s       ", "NAXIS", 2, "");
	i++; fprintf(fp, "%-8s= %20d   %-40s       ", "NAXIS1", fr->geom.sz_x, "");
	i++; fprintf(fp, "%-8s= %20d   %-40s       ", "NAXIS2", fr->geom.sz_y, "");
	if (fr->exp.datavalid) {
		i++; fprintf(fp, "%-8s= %20.3f / %-40s       ", "EXPTIME",
			     0.001 * fr->exp.exp_time, "Exposure time in seconds");
		i++; fprintf(fp, "%-8s= %20.1f / %-40s       ", "SNS_TEMP",
			     273.0 + 0.1 * fr->exp.temperature, "Temperature in K");
		i++; fprintf(fp, "%-8s= %-20s   %-40s       ", "TIMESYS", "'TT'", "");
		i++; fprintf(fp, "%-8s= %-25s   %-35s       ", "DATE-OBS", lb, 
			     "UTC OF INTEGRATION START");
		i++; fprintf(fp, "%-8s= %20.8f / %-40s       ", "JDATE",
			     jd, "JULIAN DATE OF EXP START FROM UTC");

	}


	if (!fr->fim.bsset) {
		if ((fr->stats.max - fr->stats.min) < 32767.0) {// we use positive, scaled by 1 format
			fr->fim.bscale = 1.0;
			fr->fim.bzero = 0.0;
		} else {
			fr->fim.bscale = 1.0;
			fr->fim.bzero = 32768.0;
		}
	}

	i++; fprintf(fp, "%-8s= %20.3f   %-40s       ", "BSCALE", fr->fim.bscale, "");
	i++; fprintf(fp, "%-8s= %20.3f   %-40s       ", "BZERO", fr->fim.bzero, "");

//	print out wcs data, if available
	if (fr->fim.wcsset) {
		i++; fprintf(fp, "%-8s= %20s / %-40s       ", "CTYPE1", "'RA---TAN'", "RA Projection");
		i++; fprintf(fp, "%-8s= %20.10f / %-40s       ", "CRPIX1", fr->fim.xrefpix, "");
		i++; fprintf(fp, "%-8s= %20.10f / %-40s       ", "CRVAL1", fr->fim.xref, "");
		i++; fprintf(fp, "%-8s= %20.10f / %-40s       ", "CDELT1", fr->fim.xinc, "");
		i++; fprintf(fp, "%-8s= %20.10f / %-40s       ", "CROTA1", fr->fim.rot, "N through E");
		i++; fprintf(fp, "%-8s= %20s / %-40s       ", "CTYPE2", "'DEC--TAN'", "DEC Projection");
		i++; fprintf(fp, "%-8s= %20.10f / %-40s       ", "CRPIX2", fr->fim.yrefpix, "");
		i++; fprintf(fp, "%-8s= %20.10f / %-40s       ", "CRVAL2", fr->fim.yref, "");
		i++; fprintf(fp, "%-8s= %20.10f / %-40s       ", "CDELT2", fr->fim.yinc, "");
		i++; fprintf(fp, "%-8s= %20.10f / %-40s       ", "CROTA2", fr->fim.rot, "N through E");
		i++; fprintf(fp, "%-8s= %20d / %-40s       ", "EQUINOX", fr->fim.equinox, "");
	}

// fill in /replace the noise pars

	sprintf(lb, "%20.3f / ELECTRONS/ADU", 
		fr->exp.scale);
	fits_add_keyword(fr, "ELADU", lb);
	sprintf(lb, "%20.3f / AVERAGE BIAS OF FRAME", 
		fr->exp.bias);
	fits_add_keyword(fr, "DCBIAS", lb);
	sprintf(lb, "%20.3f / READOUT NOISE", 
		fr->exp.rdnoise);
	fits_add_keyword(fr, "RDNOISE", lb);
	sprintf(lb, "%20.3f / COMBINED MULTIPLICATIVE NOISE", 
		fr->exp.flat_noise);
	fits_add_keyword(fr, "FLNOISE", lb);


// finally, print the rest of the header lines

	for (j=0; j < fr->nvar; j++) {
		for (k=0; k<FITS_HCOLS; k++)
			fputc(fr->var[j][k], fp);
		i++;
	}
		
	i++; fprintf(fp, "%-8s  %20s   %-40s       ", "END", "", "");

	k = FITS_HROWS * (i / FITS_HROWS) - i;
	if (k < 0) 
		k += FITS_HROWS;

	for (j = 0; j < k; j++)
		fprintf(fp, "%80s", "");
//	d3_printf("i=%d j=%d", i, j);

	if (fr->geom.pix_size != 4 || fr->geom.pix_format != 0) {
		err_printf("\nwrite_fits_frame: I can only write float frames\n");
		return ERR_FATAL;
	}

	all = fr->geom.sz_x * fr->geom.sz_y;
	dp = (float *)(fr->dat);

//real value = bzero + bscale * <array_value>

	for (i = 0; i < all; i ++) {
		v = floor( (dp[i] - fr->fim.bzero) / fr->fim.bscale );
		if (v < -32768)
			v = -32768;
		if (v > 32767)
			v = 32767;
		putc((v >> 8) & 0xff, fp);
		putc((v) & 0xff, fp);
	}

/* pad the end of record */
	for (i = 0;  i < 2880 - (all * 2) % 2880; i++)		
		putc(' ', fp);
	fflush(fp);
	fclose(fp);
	return 0;
}

// flat_frame divides ff by (fr1 / cavg(fr1) ; the two frames are aligned according to their skips
// the size of fr is not changed

int flat_frame(struct ccd_frame *fr, struct ccd_frame *fr1)
{
	int xovlap, yovlap;
	int xst, yst;
	int x1st, y1st;
	float *dp, *dp1;
	int x, y;
	double mu, ll;

	if (!fr1->stats.statsok)
		frame_stats(fr1);
	mu = fr1->stats.cavg;
	if (mu <= 0.0) {
		err_printf("flat frame has negative avg: %.2f aborting\n", mu);
		return -1;
	}

	if (fabs(fr->exp.bias) > 2.0) {
		err_printf("I refuse to flat a frame with non-null bias (%.f)\n", fr->exp.bias);
		return -1;
	}
	if (fabs(fr1->exp.bias) > 2.0) {
		err_printf("I refuse to use a flat with non-null bias (%.f)\n", fr1->exp.bias);
		return -1;
	}

	ll = mu / MAX_FLAT_GAIN;

	xst = fr1->geom.x_skip - fr->geom.x_skip;
	if (fr1->geom.x_skip + fr1->geom.sz_x >= fr->geom.sz_x + fr->geom.x_skip)
		xovlap = fr->geom.x_skip + fr->geom.sz_x - fr1->geom.x_skip;
	else
		xovlap = fr1->geom.x_skip + fr1->geom.sz_x - fr->geom.x_skip;
	if (xst < 0) {
		x1st = -xst; 
		xst = 0;
	} else 
		x1st = 0;

	yst = fr1->geom.y_skip - fr->geom.y_skip;
	if (fr1->geom.y_skip + fr1->geom.sz_y >= fr->geom.sz_y + fr->geom.y_skip)
		yovlap = fr->geom.y_skip + fr->geom.sz_y - fr1->geom.y_skip;
	else
		yovlap = fr1->geom.y_skip + fr1->geom.sz_y - fr->geom.y_skip;
	if (yst < 0) {
		y1st = -yst; 
		yst = 0;
	} else 
		y1st = 0;

//	d3_printf("xst %d x1st %d yst %d x1st %d xovlap %d yovlap %d",
//		  xst, x1st, yst, y1st, xovlap, yovlap);

	dp = (float *)(fr->dat) + xst + yst * fr->geom.sz_x;
	dp1 = (float *)(fr1->dat) + x1st + y1st * fr1->geom.sz_x;

	for (y = 0; y < yovlap; y++) {
		for (x = 0; x < xovlap; x++) { 
			if (*dp1 > ll)  
				*dp = *dp / *dp1 * mu;
			else 
				*dp = *dp * MAX_FLAT_GAIN;
			dp ++;
			dp1 ++;
		}
		dp += fr->geom.sz_x - xovlap;
		dp1 += fr1->geom.sz_x - xovlap;
	}
	fr->stats.statsok = 0;

	fr->exp.flat_noise = sqrt( sqr(fr1->exp.rdnoise) + mu / sqrt(fr1->exp.scale) ) / mu;

// TODO calculate noise on flat!!!
// take care of biases
	return 0;
}


// add_frames adds fr1 to fr; the two frames are aligned according to their skips
// the size of fr is not changed

int add_frames (struct ccd_frame *fr, struct ccd_frame *fr1)
{
	int xovlap, yovlap;
	int xst, yst;
	int x1st, y1st;
	float *dp, *dp1;
	int x, y;

	xst = fr1->geom.x_skip - fr->geom.x_skip;
	if (fr1->geom.x_skip + fr1->geom.sz_x >= fr->geom.sz_x + fr->geom.x_skip)
		xovlap = fr->geom.x_skip + fr->geom.sz_x - fr1->geom.x_skip;
	else
		xovlap = fr1->geom.x_skip + fr1->geom.sz_x - fr->geom.x_skip;
	if (xst < 0) {
		x1st = -xst; 
		xst = 0;
	} else 
		x1st = 0;

	yst = fr1->geom.y_skip - fr->geom.y_skip;
	if (fr1->geom.y_skip + fr1->geom.sz_y >= fr->geom.sz_y + fr->geom.y_skip)
		yovlap = fr->geom.y_skip + fr->geom.sz_y - fr1->geom.y_skip;
	else
		yovlap = fr1->geom.y_skip + fr1->geom.sz_y - fr->geom.y_skip;
	if (yst < 0) {
		y1st = -yst; 
		yst = 0;
	} else 
		y1st = 0;

//	d3_printf("xst %d x1st %d yst %d x1st %d xovlap %d yovlap %d",
//		  xst, x1st, yst, y1st, xovlap, yovlap);

	dp = (float *)(fr->dat) + xst + yst * fr->geom.sz_x;
	dp1 = (float *)(fr1->dat) + x1st + y1st * fr1->geom.sz_x;

	for (y = 0; y < yovlap; y++) {
		for (x = 0; x < xovlap; x++) { 
			*dp = *dp + *dp1;
			dp ++;
			dp1 ++;
		}
		dp += fr->geom.sz_x - xovlap;
		dp1 += fr1->geom.sz_x - xovlap;
	}

// fit noise data
	fr->exp.bias = fr->exp.bias + fr1->exp.bias;
	fr->exp.rdnoise = sqrt(sqr(fr->exp.rdnoise) + sqr(fr1->exp.rdnoise));
	fr->stats.statsok = 0;
	return 0;
}

// sub_frames substracts fr1 from fr; the two frames are aligned according to their skips
// the size of fr is not changed; fr1 is assumed to be a dark frame

int sub_frames (struct ccd_frame *fr, struct ccd_frame *fr1)
{
	int xovlap, yovlap;
	int xst, yst;
	int x1st, y1st;
	float *dp, *dp1;
	int x, y;
	double dkl;

	d3_printf("read noise is: %.1f %.1f\n", fr1->exp.rdnoise, fr->exp.rdnoise);

	xst = fr1->geom.x_skip - fr->geom.x_skip;
	if (fr1->geom.x_skip + fr1->geom.sz_x >= fr->geom.sz_x + fr->geom.x_skip)
		xovlap = fr->geom.x_skip + fr->geom.sz_x - fr1->geom.x_skip;
	else
		xovlap = fr1->geom.x_skip + fr1->geom.sz_x - fr->geom.x_skip;
	if (xst < 0) {
		x1st = -xst; 
		xst = 0;
	} else 
		x1st = 0;

	yst = fr1->geom.y_skip - fr->geom.y_skip;
	if (fr1->geom.y_skip + fr1->geom.sz_y >= fr->geom.sz_y + fr->geom.y_skip)
		yovlap = fr->geom.y_skip + fr->geom.sz_y - fr1->geom.y_skip;
	else
		yovlap = fr1->geom.y_skip + fr1->geom.sz_y - fr->geom.y_skip;
	if (yst < 0) {
		y1st = -yst; 
		yst = 0;
	} else 
		y1st = 0;

//	d3_printf("xst %d x1st %d yst %d x1st %d xovlap %d yovlap %d",
//		  xst, x1st, yst, y1st, xovlap, yovlap);

	dp = (float *)(fr->dat) + xst + yst * fr->geom.sz_x;
	dp1 = (float *)(fr1->dat) + x1st + y1st * fr1->geom.sz_x;

	for (y = 0; y < yovlap; y++) {
		for (x = 0; x < xovlap; x++) { 
			*dp = *dp - *dp1;
			dp ++;
			dp1 ++;
		}
		dp += fr->geom.sz_x - xovlap;
		dp1 += fr1->geom.sz_x - xovlap;
	}

// compute noise data
	fr->exp.bias = fr->exp.bias - fr1->exp.bias;
	fr->exp.rdnoise = sqrt(sqr(fr->exp.rdnoise) + sqr(fr1->exp.rdnoise));
	fr->stats.statsok = 0;
//	d3_printf("read noise is: %.1f %.1f\n", fr1->exp.rdnoise, fr->exp.rdnoise);
	return 0;
}




// make fr <= fr * m + s

int scale_shift_frame(struct ccd_frame *fr, double m, double s)
{
	int i, all;
	float *dp;

	all = fr->geom.sz_x * fr->geom.sz_y;
	dp = (float *)(fr->dat);

	for (i=0; i<all; i++) {
		*dp = *dp * m + s;
		dp ++;
	}
	fr->exp.bias = fr->exp.bias * m + s;
	fr->exp.scale /= abs(m);
	fr->exp.rdnoise *= abs(m);
	fr->stats.statsok = 0;
	return 0;
}

// lookup a keyword in the frame's vars. Return a pointer to it's
// line, or NULL if not found
FITS_row *fits_keyword_lookup(struct ccd_frame *fr, char *kwd)
{
	FITS_row *var;
	int i;

	var = fr->var;
	for (i=0; i<fr->nvar; i++) {
		if (!strncmp((char *)(var + i), kwd, strlen(kwd)))
			return (var + i);
	}
	return NULL;
}

// add a string keyword to the fits header
// quotes are _not_ added to the string
// if the keyword exists, it is replaced
// return 0 if ok
int fits_add_keyword(struct ccd_frame *fr, char *kwd, char *val)
{
	char lb[81];
	FITS_row *v1;

	sprintf(lb, "%-8s= %-70s", kwd, val);
	v1 = fits_keyword_lookup(fr, kwd);
	if (v1 != NULL) { // replace current value
		memcpy(v1, lb, sizeof(FITS_row));
	} else { // alloc space for a new one
		v1 = realloc(fr->var, (fr->nvar + 1) * FITS_HCOLS);
		if (v1 == NULL) {
			err_printf("cannot alloc mem for keyword %s\n", kwd);
			return ERR_ALLOC;
		} 
		fr->var = v1;
		memcpy(fr->nvar + v1, lb, FITS_HCOLS);
		fr->nvar ++;
	}
	return 0;

}

// append a HISTORY keyword to the fits header
// quotes are _not_ added to the string
// return 0 if ok
int fits_add_history(struct ccd_frame *fr, char *val)
{
	char lb[81];
	FITS_row *v1;

	sprintf(lb, "HISTORY = %-70s", val);
	// alloc space for a new one
	v1 = realloc(fr->var, (fr->nvar + 1) * FITS_HCOLS);
	if (v1 == NULL) {
		err_printf("cannot alloc mem for history \n");
		return ERR_ALLOC;
	} 
	fr->var = v1;
	memcpy(fr->nvar + v1, lb, FITS_HCOLS);
	fr->nvar ++;
	return 0;
}

// crop_frame reduces the size of a frame; the upper-left corner of the frame will
// added to the current skip. The data area is realloced. Returns non-zero in case of an error.

int crop_frame(struct ccd_frame *fr, int x, int y, int w, int h)
{
	float *sp, *dp;
	int ix, iy;
	void *ret;

	d3_printf("crop from %d %d size %d %d \n", x, y, w, h);

// check that the subframe is fully contained in the source frame
	if (x < 0 || x + w > fr->geom.sz_x || y < 0 || y + h > fr->geom.sz_y) {
		err_printf("crop_frame: bad subframe\n");
		return ERR_FATAL;
	}
	sp = dp = (float *)(fr->dat);
	sp += x + y * fr->geom.sz_x;

// copy the contents to the upper-left corner
	for (iy=0; iy<h; iy++) {
		for(ix=0; ix<w; ix++) {
			*dp++ = *sp++;
		}
		sp += fr->geom.sz_x - w;
	}
	ret = realloc(fr->dat, sizeof(float)*w*h);
	if (ret == NULL) {
		err_printf("crop_frame: alloc error \n");
		return ERR_ALLOC;
	}
// adjust the frame info
	fr->geom.x_skip += x;
	fr->geom.y_skip += y;
	fr->geom.sz_x = w;
	fr->geom.sz_y = h;
	fr->dat = ret;
	fr->stats.statsok = 0;
	return 0;
}

