;;;; ;; GCXCHED a scheduler of variable star observations for GCX driven
;;;; ;;         cameras and telescopes
;;;; ;; Copyright (c) 2003 Alexandru Dan Corlan MD PhD
;;;; ;;

;;;; ;;    This program is free software; you can redistribute it and/or modify
;;;; ;;    it under the terms of the GNU General Public License as published by
;;;; ;;    the Free Software Foundation; either version 2 of the License, or
;;;; ;;    (at your option) any later version.

;;;; ;;    This program is distributed in the hope that it will be useful,
;;;; ;;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; ;;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; ;;    GNU General Public License for more details.

;;;; ;;    You should have received a copy of the GNU General Public License
;;;; ;;    along with this program; if not, write to the Free Software
;;;; ;;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(require "AVSOBS" "/home/alexandru/proj/avsomat/src/avsobs.lisp")

;;;; selectors of lists of stars

(defun in-constellation (const &optional (oflist *vs*))

  "Return the variable star symbols of objects in constelation CONST
   from the list OFLIST (all variable stars in GCVS by default)."
  
  (mapcan #'(lambda (x)
	      (let ((starname (symbol-name x))
		    (constname (symbol-name const)))
		(if (string= starname constname :start1 (- (length starname) 3))
		    (list x)
		  nil)))
	  oflist)
  )

(compile 'in-constellation)

(defun having-aavso-chart (starlist)

  "Select from starlist (a list of variable star symbols) those
   that also have an aavso chart"
  
  (mapcan #'(lambda (x)
	      (if (memq x *with-aavso-charts*)
		  (list x)
		nil))
	  starlist)
  )

(defun eruptive (starlist)

  "Extract from the STARLIST list of variable star symbols those
   of a type which could make them to erupt repeatedly."

  (mapcan #'(lambda (x) 
	      (let ((vt (get x 'vartype)))
		(if (and (stringp vt)
			 (or (search "FU" vt)
			     (search "GCAS" vt)
			     (string= "I" vt :end2 1)
			     (search "RCB" vt)
			     (search "SB" vt)
			     (search "SDOR" vt)
			     (search "UV" vt)
			     (search "UVN" vt)
			     (search "WR" vt)
			     (search "UG" vt)
			     (search "UGSS" vt)
			     (search "UGSU" vt)
			     (search "UGZ" vt)
			     (search "ZAND" vt)
			     (search "BLLAC" vt)
			     ))
		    (list x)
		  nil)))
	      starlist)
	  )



;;;; Optimizer of the path followed by the telescope
;;;; to visit the stars in a list

(defun sort-by-ra (starlist)

  "Sort list of variable star symbols by right ascension."
  
  (sort starlist #'(lambda (x y)
		     (< (get x 'ra) (get y 'ra))
		     )
	)
  )

(defun real-cart-dist (ra1 dec1 ra2 dec2)

  "Calculate a distance between star1 and star2 (given by coordinates
   in degrees) in terms of cartezian distance between the two coordinates
   (does not work well in the vicinity of the pole)."

  (declare (single-float ra1)
	   (single-float dec1)
	   (single-float ra2)
	   (single-float dec2))
  (let ((dra (the single-float (- ra1 ra2)))
	(ddec (the single-float (- dec1 dec2))))
    (the single-float
      (sqrt (the single-float
	      (+ (the single-float (* dra dra))
		 (the single-float (* ddec ddec))))))
    )
  )

(defun cart-dist (ra1 dec1 ra2 dec2)

  "Calculate a distance between star1 and star2 (given by coordinates
   in degrees) proportional to the time taken for the telescope to go
   from star1 to star2 (present version may not work well in the
   vicinity of the pole)."

  (declare (single-float ra1)
	   (single-float dec1)
	   (single-float ra2)
	   (single-float dec2))
  (let* ((dra (the single-float (abs (the single-float (- ra1 ra2)))))
	 (ddec (the single-float (abs (the single-float (- dec1 dec2)))))
	 )
    (the single-float (max dra ddec))
    )
  )

(compile 'cart-dist)

(defun star-cart-dist (star1 star2)

  "Calculate a distance between star1 and star2 (given by symbolic
   names) in terms of time taken for the telescope to go
   from star1 to star2 (present version may not work well in the
   vicinity of the pole)."

  (cart-dist (coerce (get star1 'ra) 'single-float)
	     (coerce (get star1 'dec) 'single-float)
	     (coerce (get star2 'ra) 'single-float)
	     (coerce (get star2 'dec) 'single-float)))

(compile 'star-cart-dist)

(defun star-path-length (starlist)

  "Compute the length in degrees of the path needed to visit
   all stars named in the STARLIST."
  
  (let ((ll 0.0))
    (if (listp starlist)
	(maplist #'(lambda (x) (if (cdr x) (incf ll (star-cart-dist (car x) (cadr x)))))
		 starlist)
      )
    (if (vectorp starlist)
	(dotimes (i (length starlist))
	  (if (> i 0) (incf ll (star-cart-dist (elt starlist (1- i))
					       (elt starlist i)))))
      )
    ll))

(compile 'star-path-length)

(defun depict-star-path (starlist filename)

  "Given a list of variable star symbols, produce a postscript
   file with the path a telescope would make to visit all these stars."

  (with-open-file (so filename :direction :output :if-exists :supersede)
		  (let ((xo (get (car starlist) 'ra))
			(yo (get (car starlist) 'dec)))
		    (format so "300 400 translate 5 5 scale~%")
		    (format so "0.1 setlinewidth~%")
		    (format so "newpath 0.0 0.0 moveto~%")
		    (dolist (nestar (cdr starlist))
		      (format so "~A ~A lineto~%"
			      (- (get nestar 'ra) xo)
			      (- (get nestar 'dec) yo))
		      )
		    (format so "stroke showpage quit~%~%")
		    )
		  )
  )

(defun optimize-post-by-ra (ralist)

  "Assume RALIST is a list of variable star symbols sorted by
   right ascension (as done by SORT-BY-RA). Swap pairs of stars
   if this reduces the path made by the telescope to visit all of them."
  
  (let* ((stl (coerce ralist 'vector))
	 (l (length stl))
	 (crtpl (star-path-length stl))
	 (npl crtpl)
	 )
    (if (> l 7)
	(dotimes (i (- l 5))
	  (dotimes (j (- l i 1))
	    (psetf (elt stl i) (elt stl (+ i j 1))
		   (elt stl (+ i j 1)) (elt stl i))
	    (setf npl (star-path-length stl))
	    (if (< npl crtpl)
		(setf crtpl npl)
	      (psetf (elt stl i) (elt stl (+ i j 1))
		     (elt stl (+ i j 1)) (elt stl i)))
	    )
	  )
      )
    stl
    )      
  )

(compile 'optimize-post-by-ra)

(defun sort-by-optimal-path (starlist)

  "One optimum path optimizer. Take all pairs of stars, sort by
   distance between them, then try to use the closest pairs to build
   up a path. Not used in the final solution."

  (defstruct starpair
    (elti 'fixnum)
    (eltj 'fixnum)
    (dist 'single-float))
  
  (let* ((l (length starlist))
	 (ras (make-array l :element-type 'single-float))
	 (decs (make-array l :element-type 'single-float))
	 (follow (make-sequence '(vector t) l :initial-element nil))
	 (dists (make-sequence '(vector starpair) (/ (* l (1- l)) 2)))
	 (di 0)
	    ;	 (d (make-array (list l l) :element-type 'single-float))
	 )
    (map-into ras #'(lambda (x) (coerce (get x 'ra) 'single-float)) starlist)
    (map-into decs #'(lambda (x) (coerce (get x 'dec) 'single-float)) starlist)
    (dotimes (i l)
      (dotimes (j l)
	(if (< i j)
	    (progn
	      (setf (elt dists di)
		    (make-starpair
		     :elti i
		     :eltj j
		     :dist (cart-dist (aref ras i)
				      (aref decs i)
				      (aref ras j)
				      (aref decs j))
		     )
		    )
	      (incf di))
	  )
	)
      )
    (sort dists #'(lambda (x y) (< (starpair-dist x) (starpair-dist y))))
    (let ((entr (starpair-elti (elt dists 0)))
	  (retl (make-sequence '(vector t) l)))
      (map nil #'(lambda (x)
		   (if (and (not (elt follow (starpair-elti x)))
			    (not (find (starpair-eltj x) follow))
			    (not (eql (starpair-eltj x) entr)))
		       (setf (elt follow (starpair-elti x)) (starpair-eltj x))
		     (if (and (not (elt follow (starpair-eltj x)))
			      (not (find (starpair-elti x) follow))
			      (not (eql (starpair-elti x) entr)))
		       (setf (elt follow (starpair-eltj x)) (starpair-elti x))))
		   )
	   dists)
;;      (format *standard-output* "~A~%~A~%" follow
;;	      (sort (copy-seq follow) #'(lambda (x y) (if (or (not x) (not y)) t (< x y)))))
      (dotimes (i l)
	(setf (elt retl i) (elt starlist entr))
	(psetf entr (elt follow entr)
	      (elt follow entr) nil)
	(if (not entr)
	    (if (position-if #'(lambda (x) x) follow)
		(progn
		  (setf entr (elt follow (position-if #'(lambda (x) x) follow)))
		  (setf (elt follow (position-if #'(lambda (x) x) follow)) nil)
		  )
	      )
	  )
	)
      retl)
    )
)
  
(compile 'sort-by-optimal-path)

(defun report-cart-path-lengths (const)

  "Print a table of the path lengths obtained with the various methods
   of optimization described above."

  (let ((inconst (having-aavso-chart (in-constellation const))))
    (if (> (length inconst) 3)
	(format *standard-output* "~A ~6@s ~10,2f ~10,2f ~10,2f ~10,2f ~10,2f~%"
		const
		(length inconst)
		(star-path-length inconst) 
		(star-path-length (sort-by-ra (copy-seq inconst)))
		(star-path-length (optimize-post-by-ra (sort-by-ra (copy-seq inconst))))
		(star-path-length (optimize-post-by-ra
				   (optimize-post-by-ra (sort-by-ra (copy-seq inconst)))))
		(star-path-length (sort-by-optimal-path (copy-seq inconst)))
		)
      )
    )
)


(defun depict-ora (const fn)

  "Given a constellation, produce a postscript
   file with the path a telescope would make to visit all the stars
   in the constellation having aavso charts."

  (depict-star-path
   (coerce (optimize-post-by-ra
	    (optimize-post-by-ra
	     (sort-by-ra
	      (having-aavso-chart (in-constellation const))))) 'list) fn)
  )

;;;; Producing scripts for the gcx-controlled telescope to follow

(defun undash (starname) ;;; ! ineficient version !

  "Remove the dash from the symbol-name of the star and
   return the variable star name as a string."

  (concatenate 'string
	       (subseq (symbol-name starname) 0 (- (length (symbol-name starname)) 4))
	       (subseq (symbol-name starname) (- (length (symbol-name starname)) 3)
		       (length (symbol-name starname)))))

		  

(defun script-star (star)

  "Generate the script segment to drive the gcx utility to
   picture the star."

  (format nil "# ~A (~A) ~A ~A ~A ~A~%goto ~A~%get~%match~%mget 2~%"
	  star (get star 'vartype) (get star 'ra) (get star 'dec)
          (if (listp (get star 'mag))
	      (if (listp (cadar (get star 'mag)))
		  (format nil "~Amag=~A-~A"
			  (caar (get star 'mag))
			  (caadar (get star 'mag))
			  (cadr (cadar (get star 'mag))))
		"mag=?")
	    "mag=?")
          (if (memq star *rcps*) "+rcp+" "")
	  (string-downcase (undash star))))

(defun quick-script-star (star)

  "Generate the script segment to drive the gcx utility to
   picture the star, without retaking the positioning picture
   unless necessary."

  (format nil "# ~A (~A) ~A ~A ~A ~A~%goto ~A~%get~%ckpoint~%save~%get~%save~%"
	  star (get star 'vartype) (get star 'ra) (get star 'dec)
          (if (listp (get star 'mag))
	      (if (listp (cadar (get star 'mag)))
		  (format nil "~Amag=~A-~A"
			  (caar (get star 'mag))
			  (caadar (get star 'mag))
			  (cadr (cadar (get star 'mag))))
		"mag=?")
	    "mag=?")
          (if (memq star *rcps*) "+rcp+" "")
	  (string-downcase (undash star))))

(defun script-eruptive-in (constel)

  "Generate a gcx script for the eruptive stars in constellation CONSTEL."

  (apply #'concatenate 'string
	 (mapcar #'script-star
		 (sort-by-ra (having-aavso-chart (eruptive (in-constellation constel)))))))

(defun script-important-in (constel)

  "Generate a gcx script for the eruptive and important LPV stars
   (stars in need of more observations) from constellation CONSTEL
   which also have an aavso chart."

  (apply #'concatenate 'string
	 (mapcar #'script-star
		 (sort-by-ra
		  (having-aavso-chart 
		   (union (eruptive (in-constellation constel))
			  (in-constellation constel *needmore*)))))))

(defun script-aavso-in (constel)

  "Generate a gcx script for all the stars in constellation CONSTEL
   which have an aavso chart."

  (apply #'concatenate 'string
	 (mapcar #'script-star
		 (sort-by-ra (having-aavso-chart (in-constellation constel))))))

(defun script-all-aavso-eruptives ()
  
  "Generate a gcx script for the eruptive and important LPV stars
   (stars in need of more observations) from constellation CONSTEL."

  (mapcar #'(lambda (x)
	      (format *standard-output* "echo \"~%~A\" >erup~A.obs~%~%"
		      (script-eruptive-in x)
		      (string-downcase x)))
	  *constellations*))

(defun script-all-aavso-important ()

  "Generate a gcx script for all the variable stars in need of more observations
   and those that may erupt of those that have an aavso chart."

  (mapcar #'(lambda (x)
	      (format *standard-output* "echo \"~%~A\" >imp~A.obs~%~%"
		      (script-important-in x)
		      (string-downcase x)))
	  *constellations*))

(defun script-all-aavso ()

  "Generate a gcx script to image all stars which have an aavso chart
   sorted by constellation."

  (mapcar #'(lambda (x)
	      (format *standard-output* "echo \"~%~A\" >~A.obs~%~%"
		      (script-aavso-in x)
		      (string-downcase x)))
	  *constellations*))

(defvar memo-eruptive nil)

(defun gcx-script (expr &optional (with #'script-star))
  "Produce a GCX script to image the stars designated by the expression,
   following an optimised path and going from west to east. Function
   WITH will be used to produce the script of each star designated
   by EXPR. It must return a string given a symbol.

   By variable stars, below, we understand the stars in the GCVS.lisp
   file, which was initially based on GCVS.

   EXPR may be a symbol, which could be:
     - a star name (example: RU-LYN)
     - a constellation TLA (three letter abbreviation)--variable stars
       in that constellation (example: LYN)
     - :IN-AAVSO stars having an aavso chart
     - :ERUPTIVE stars with an `eruptive' type of variability
     - :NEEDMORE stars `in need of more observations'

   If EXPR is a list, then it must have one of the following forms:

     (THE STARNAME*)    - an explicit list of stars
     (AND EXPR*)        - stars designated simultaneously by all the EXPR's
     (OR EXPR*)         - any of the stars designated by an expr
     (ANDNOT EXPR EXPR) - stars designated by the first EXPR without
                        - those designated by the second EXPR

   Examples:

   (GCX-SCRIPT 'RU-LYN) ; returns a script to image RU-LYN
   (GCX-SCRIPT 'LYN)    ; returns a script to image all variable stars known in LYN
   (GCX-SCRIPT 'LYN #'(lambda (star) (format nil \"~A~%\" star)))
                        ; returns a string with the variable stars in LYN
                        ; sorted aproximately from W to E and with optimisation
                        ; of the path between them
   (GCX-SCRIPT '(AND LYN :IN-AAVSO))
                        ; returns a script to image all stars in LYN having aavso charts
   (GCX-SCRIPT '(AND :ERUPTIVE LYN))
                        ; returns a script to image all eruptive stars in LYN
   (GCX-SCRIPT '(ANDNOT (AND :ERUPTIVE LYN) :IN-AAVSO))
                        ; returns a script to image all eruptive star not in aavso
   "

   (defun star-list (expr)
     (cond
      ((symbolp expr)
       (cond
	((member expr *constellations*) (in-constellation expr))
	((eq expr :in-aavso) *with-aavso-charts*)
	((eq expr :eruptive)
	 (if (not memo-eruptive)
	     (setf memo-eruptive (eruptive *vs*)))
	 memo-eruptive
	 )
	((eq expr :needmore) *needmore*)
	(t (list expr))))
      ((listp expr)
       (cond
	((eq (car expr) 'the)
	 (cdr expr))
	((eq (car expr) 'and)
	 (apply #'intersection (mapcar #'star-list (cdr expr))))
	((eq (car expr) 'or)
	 (apply #'union (mapcar #'star-list (cdr expr))))
	((eq (car expr) 'andnot)
	 (apply #'set-difference ;; error cases must be dealt with here
		(star-list (cadr expr))
		(star-list (cons 'or (cddr expr)))))
	))
			   
      )
     )

   (apply #'concatenate
	  (cons 'string
		(mapcar with
			(coerce
			 (optimize-post-by-ra
			  (optimize-post-by-ra
			   (sort-by-ra (star-list expr))))
			 'list)
			))
	  )
	  
)


(defun scripts ()
  "Remake all scripts."
;  (mapcar #'(lambda (constl)
;	      (with-open-file (fo (format nil "~A/gcxscript/~A.obs"
;					  *avsomat-home* constl)
;				  :direction :output :if-exists :supersede)
;		(format fo "~A"
;		  (gcx-script (list 'and constl :in-aavso)
;			      #'quick-script-star))))
;	  *constellations*)
  (mapcar #'(lambda (constl)
	      (with-open-file (fo (format nil "~A/gcxscript/erup~A.obs"
					  *avsomat-home* constl)
				  :direction :output :if-exists :supersede)
		(format fo "~A"
		  (gcx-script (list 'and constl (and :in-aavso :eruptive))
			      #'quick-script-star))))
	  *constellations*)
  (mapcar #'(lambda (constl)
	      (with-open-file (fo (format nil "~A/gcxscript/imp~A.obs"
					  *avsomat-home* constl)
				  :direction :output :if-exists :supersede)
		(format fo "~A"
		  (gcx-script (list 'and constl (and :in-aavso :needmore))
			      #'quick-script-star))))
	  *constellations*)
  )

(compile 'gcx-script)
(compile 'quick-script-star)
(compile 'scripts)
