// Libnut---utilities for starry CCD images processing
// Copyright (c) 2003 Alexandru Dan Corlan (http://dan.corlan.net)
//
// This file is the header and the documentation of the library

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

// SECTION 1.
//
// STARMATCH
// Recognition of star patterns, presumably extracted from catalogs or
// CCD images; sx,sy are centers, in pixel units, of star-like sources presumably on a
// CCD image while cx,cy are centers, in pixel units, of star-like sources, presumably
// in a catalog or also on a CCD image; sm are magnitudes (relative
// to each other) on the plate, cm are magnitudes (relative to each other)
// on the catalog or another plate; sn and cn are the numbers of source
// in the two sets to match; cix is a vector of indexes in c, one for
// each source in s in which a match was found;
// for example, if s[5] corresponds to c[67] then cix[5]==67
// non-matches are indicated by a negative number
// omega is the angle (in radians)
// to apply to c sources, around 0 to match those in s; dx,dy are
// displacements of the c sources, in pixel units, to match sources in s
// dmag is the magnitude to add to sm[i] to obtain the catalog value
// (average estimate of differences between matching pairs); ampx, ampy
// are (very close to 1.0) scalings on x and y to optimize matching with
// sources
//
// returns: the number of matching pairs or a negative error code + out values
// collateral effects: none
// reentrant: no

int starmatch(float *sx, float *sy, float *sm, int sn,    // input sources in picture
              float *cx, float *cy, float *cm, int cn,    // input catalog
              int *cix, float *omega, float *dx, float *dy, 
              float *ampx, float *ampy,
              float *dmag);  // out; cix is a vector [sn]

void set_amplif_tolerance(float mi, float ma); // default 0.85, 1.15
void set_mag_tolerance(float mi, float ma);    // default -1.5, +1.5

// TRANSCAT
//
// Transform the ``catalog'' dataset (c) presumably to match a plate;
// values in cx,cy are changed to transformed ones
//
// int transcat(float *cx, float *cy, int cn,
//              float omega, float dx, float dy, float ampx, float ampy);
//
// SECTION 2. VARIABLE STAR RECIPES
//
// The term `recipe' is used below in a broad sense: a heterogenous description
// of the objects in a certain starfield and in a narrow sense: a set
// of stars which are to be used as reference stars when performing 
// photometry of an object of interest. The term `shorcat' is used 
// in the first case when a confusion is possible.
//
// Recipes for identifying interesting starfields (primarily starfields
// around a variable star in certain bands); recipes will be represented
// a ascii files, each containing a set of objects; each object fills
// an integer number of lines; in the first line, in the first column,
// the type of the object (CONSTAR or VARSTAR) is stored; the following
// lines which start with a space belong to the same object
// three unnamed fields (local-id, ra, dec) follow, separated by spaces,
// follow. Local-id must be a contiguous set of non-space characters.
// It must be unique in the starfield or be the character `*' which
// means an id must be assigned by the system.
//
// Following constructions are of the form NAME=VALUE
// where the value is either a contiguous string of non-space
// characters or a string which may contain spaces but is
// included in balanced '()'; in the later case it usually contains
// a list of elements; the meaningful names of fields are listed below;
// non-meaningful names are ignored; stared names may be repeated
//
// IMAG -- magnitude on the whole ccd spectrum; a fractional number or
//      -- an expression which can be (min max) to specify a range
//      -- or a more complex model, not implemented yet and not described here
// PMAG -- photographic magnitude from gsc
// VMAG, BMAG, RMAG, IRMAG -- magnitudes in informal certain bands
// VTMAG, BTMAG -- magnitudes from the Tycho-2 catalog
// FGMAG, JGMAG, VGMAG, NGMAG -- magnitudes from gsc2
// USE-V-REF -- a list of other stars from the recipe to use a reference when
//        -- doing differential photometry, eventually with a range of
//        -- magnitudes over which the reference is recommended; for example
//        -- (A B F) to use stars labeled `A', `B' and `F' or 
//        -- (9.0 12.0 A C D) to use `A', `C' and `D' as references for exact
//        -- photometry in V band when the object has a magnitude between 9.0 and 12.0
// USE-B-REF USE-U-REF USE-IR-REF etc
//        -- for other bands
// SPECTYPE -- spectral type
// VARTYPE -- variable type
// VARID  -- identifier from the aavso validation file (for variable stars)
// EPOCH -- epoch at which the object positions are given (not supported yet)
// GSCL -- linkout to gsc ex: GSCL=2176-0580
// GSC2L -- linkout to gsc2
// PERIOD -- period in days (not supported yet)
// 
// CMT -- comment (not supported yet)
//
// Example: vwvul.rcp

// Functional interface.
//
// Two enumerated types

#define NO_STAR 0                        // type of object in the shorcat
#define VARSTAR 1
#define CONSTAR 2

#define P_BAND      0                    // colocvial names for spectral bands
#define B_BAND      1
#define V_BAND      2
#define U_BAND      3
#define R_BAND      4
#define VT_BAND     5
#define BT_BAND     6
#define IR_BAND     7
#define VG_BAND     8
#define FG_BAND     9
#define JG_BAND    10
#define NG_BAND    11

#define N_BANDS    12

// Number of stars in the shorcat. Shorcats are text files, identified
// by filename. The library can be assumed to maintain a (parsed) buffer of one
// shorcat. If repeated calls from the functions below are issued 
// for the same catalog name the buffer is not reread and reparsed from disk.

// Objects in the shorcat can be identified by integers up to the number returned
// these integer are called `star indices' below

// Returns a negative number for any error

int rcipe_n_stars(char *catname);

// The following three functions return the sets of coordinates
// and magnitudes of the stars in the recipe (short catalog)
// in the out_ vectors which must have been preallocated (typically
// of rcipe_n_stars() size). The functions below return the number of 
// objects actually stored or a negative number for error.
// rcipe_xy is like rcipe_coordinates but the positions of objects
// are projected using the `TAN' projection, ready to be directly
// starmatch-ed aganinst sources extracted from an image.

// band is one of the _BAND defines above

int rcipe_coordinates(char *catname, float *out_ra, float *out_dec, int size);
int rcipe_xy(char *catname, float *out_x, float *out_y, float deg_per_pix, int size);
int rcipe_magnitudes(char *catname, int band, float *out_m, int size);

// Store, in out_r, the star indices to be used as exact photometry references
// for the (presumably variable) star with index `star' if it has
// the approximate magnitude mg in the band (which is one of the _BAND
// defines above). out_r must be preallocated and its size is `size'.

int rcipe_refs(char *catname, int star, float mg, int band, int *out_r, int size);

// Store star indices of the variable and constant stars from shorcat
// in rstars (which must have been preallocated by the caller). 
// Size is the allocated length of rstars to avoid overflows.
// Number of retrieved star indices is returned, or a negative for error.

int rcipe_varstars(char *catname, int *rstars, int size);
int rcipe_constars(char *catname, int *rstars, int size);

// Return various parameters of an object given by star index `star'
// from the shorcat. A value <= -1000.0 is returned
// in case of error. 

float rcipe_ra(char *catname, int star);
float rcipe_dec(char *catname, int star);
float rcipe_mag(char *catname, int star, int band);
float rcipe_highest_mag(char *catname, int band, int star);

// similar to the above, but the type (NO_STAR/VARSTAR/CONSTAR) is returned
// for the star with index `star'. 

int rcipe_type(char *catname, int star);

// A string component of an object is copied (including 
// trailing \0) into the user-supplied obuf which must be
// of sufficient size (100 characters will suffice),
// given as argument `size'.

int rcipe_label(char *catname, int star, char *obuf, int size);
int rcipe_gsclink(char *catname, int star, char *obuf, int size);
int rcipe_gsc2link(char *catname, int star, char *obuf, int size);
int rcipe_spectype(char *catname, int star, char *obuf, int size);
int rcipe_varid(char *catname, int star, char *obuf, int size);    // aavso code
int rcipe_vartype(char *catname, int star, char *obuf, int size);

int rcipe_chartid(char *catname, int star, char *obuf, int size);

float rcipe_postol(char *catname, int star); // position tolerance in seconds of arc
// search for the closest source withing this radius of the catalog value and, 
// if none found, report `too faint to see'


// Search for the star index with the argument label in the shorcat
// returns the star index or a negative number if error (-2 if not found)

int rcipe_named(char *catname, char *lbl);

// Call rcipe_test with the name of catalog to go through most of the
// above functions and print some data on standard output
// see also its source code for an example of using
// the above.

void rcipe_test(char *catname);

// 3. OTHER CATALOGS
//
// The following function, which is to be found in gsc/embgsc.c, reads
// data around a ra/dec position from the gsc catalogue; it needs to link to
// all other files in the gsc subdirectory

int getgsc(float ra, float dec, float radius, float maxmag, int *regs,
	   int *ids, float *ras, float *decs, float *mags, int n);

// 4. Extract information from a fits image: header data and list of sources
//    with magnitudes relative to each other
//
//    src_x, src_y, mgns must have been already allocated with at least len floats
//    object must have been already allocated with at least object_len characters
//


int ip_lextract(char *fits_fn, float snr, int burnl,                    // input data
	        float src_x[], float src_y[], float mgns[], int len,    // source list outp
	        char object[], int object_len,                          // target output
	        double *obstime, float *xref, float *yref,              // fits head output
                float *xrefpix, float *yrefpix, float *xinc, float *yinc,
                float *rot);

float photometrisate(char *fits_fn, double tgx, double tgy,
		     double refx, double refy, double magr);
