// shorcat -- library for dealing with short catalogs of celestial objects
// Copyright (c) 2003 Alexandru Dan Corlan (http://dan.corlan.net)
//
// the term `recipe' is sometimes used for a short catalog
// here `recipe' is also used to define a method of measuring the
// magnitude of a variable star in a certain range
//
// this code is not reentrant (must be called from a single thread)

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

// UGLY HACK at line 266 or so

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <math.h>
#include "libnut.h"

#define FN_MAX 1000
#define STAR_MAX 50000              // 10_000
#define BUF_MAX 1000000             // 100_000
#define REF_MAX 20                  // max number of reference stars for a recipe
#define RECIPE_MAX 100              // of all variable stars described in the short catalog

static int initialized = 0;         // 1 if everything was allocated

static char filename[FN_MAX] = "";  // name of short catalog (the name of the file)
static char buf[BUF_MAX];           // buffer of strings to be read from the catalog
static int bufill = 0;

static float ra[STAR_MAX];
static float dec[STAR_MAX];
static float postol[STAR_MAX];      // radius, in arcsec, around the catalog position
                                    // where the star should be searched for


static char stype[STAR_MAX];        // one of the above values;
                                    // must initialize to NO_STAR (impossible to 
                                    // specify in C)

static char *chart[STAR_MAX];       // name of aavso chart
static char *label[STAR_MAX];       // pointers in buf; must initialize to NULL
static char *vartype[STAR_MAX];     // idem, type of variable star, NULL for constant
static char *spectype[STAR_MAX];    // idem, spectral type of star
static char *varid[STAR_MAX];       // idem, name of variable star, NULL for constant
static char *gsclink[STAR_MAX];     // idem, gsc id of the star
static char *gsc2link[STAR_MAX];     // idem, gsc id of the star

typedef struct recipe {
  int band;
  float lowmag;
  float himag;
  int refs[REF_MAX];
  int nrefs;                         // Can't init to 0 here! nasty bugs may happen easily!
  char *src;                         // = NULL; unprocessed source of the recipe
} RECIPE;

static RECIPE rec[RECIPE_MAX];

static float *mag[N_BANDS];         // to be malloced for supported bands
static float *magmax[N_BANDS];      // idem

static short first_recipe[STAR_MAX];// index in rec (for variable stars) or -1
static short last_recipe[STAR_MAX]; // idem

static int starfill = 0;            // for all star objects
static int recfill = 0;             // next recipe to fill

static int initialize() { // returns non-0 if malloc fails
  int i;

  if (initialized)
    return 0;     // double initialization not done, but call tolerated

  for (i=0; i<N_BANDS; i++) {
    if ((mag[i] = (float *)malloc(STAR_MAX * sizeof(float)))==NULL)
      return 1; // this, of course, lets all previous mallocs unfreed
    if ((magmax[i] = (float *)malloc(STAR_MAX * sizeof(float)))==NULL)
      return 1; // idem
  }

  initialized = 1;
  return 0;
}

static FILE *fi = NULL;

#define TOKMAX 10000    // 10_000

static char token[TOKMAX];  // to be seen simultaneously from the following functions
                            // (until read_shortcat) which should normally be local
                            // to read_shortcat

static int read_token(FILE *f) {  // return 1 if read, 0 if end of file reached

  int i;
  int c;  // I mean char, of course, but must deal with eof

  while (isspace(c=getc(f))) ;
  if (feof(f))
    return 0;  // don't forget to add accolades after if() 
               // if inserting something before this instruction
  ungetc(c,f);
  while (c==';') {
    while((c=getc(f))!='\n') ;
    while(isspace(c=getc(f))) ;
    ungetc(c,f);
  }
  for (i=0; (i<TOKMAX) && ((c=getc(f))!='=') && (!isspace(c)) && (c!=EOF); i++) {
    token[i] = c;
  } // this time I added them just to be sure
  if (i==TOKMAX)
    return 0;
  if (c=='=') {
    if (i<TOKMAX)
      token[i++] = c;
    else
      return 0; // how will the caller know this is not end-of-file but a comment
                // longer than TOKMAX or a forgotten ')' ? if somebody will want
                // to transform read_token into a trilogy he could both document
                // this and improvise an exception processing mechanism 
                // in the caller
    c = getc(f);
    if (c=='(') {
      if (i<TOKMAX)
	token[i++] = c;
      else
	return 0; // I've seen this idiom before...
      for( ; (i<TOKMAX) && ((c=getc(f))!=')' && (c!=EOF)); i++) {
	token[i] = c;
      } // accolades again
      if (i>=TOKMAX)
	return 0;  // that idiom again
      if (c==')')
	token[i++] = c;
      if (c==EOF)
	return 0;  // missed ')'; user will think file ended; should signal differently
      if (i>=TOKMAX)
	return 0;  // same idiom, sorry, must avoid buffer overflow
      else
	token[i] = 0;
      return 1;
    } else { // non '()' tokens, look for space to end token
      token[i++] = c;
      for (; i<TOKMAX && ((c=getc(f))!=EOF) && (!(isspace(c))); i++) {
	token[i] = c;
      } // the useful habit of always puting accolades in C
      if (i>=TOKMAX)
	return 0; // unlikely but possible; why `idiom' is like `idiot' ?
      else        // `idiot' comes from old-greek `idiotes'
	token[i++] = 0; // meaning ``one who talks only to himself''
      return 1;
    }
    // could in future have other cases here, so it is unwise to let
    // control `break' through the above
  }
  if (i>=TOKMAX)    
    return 0;       // `when Constraint_Error => Say_Something;' in _any_
  else              // caller who needs it would have been easier and clearer
    token[i++] = 0; // puting this idiom all over the place but ... still happy
  return 1;         // its a single buffer to care of; hope I treated all cases
}

/* testing ...
static int read_token(FILE *fi) {
  int r;

  r = _read_token(fi);
  printf("|%s",token);
  return r;
}
*/

char *stringstore(char *str) { // returns NULL if buffer full

  int l;

  l = strlen(str);
  if (l+bufill+1 >= BUF_MAX) {
    fprintf(stderr, "Fatal: buf overflow\n");
    exit(1); // if you want to recover from this exception later
  }          // add a wrapper to each function which calls 
             // {a function which calls}* stringstore
  strcpy(buf+bufill, str);
  bufill += l+1; // providing for ending \0
  return buf+bufill-l-1;
}

static void add_unparsed_recipe(int star, int band, char *tokstart) {

  int k;
  
  if (first_recipe[star]==-1)
    first_recipe[star] = recfill;
  rec[recfill].src = stringstore(token+strlen(tokstart));
  rec[recfill].nrefs = 0;
  for (k=1; k<REF_MAX; k++)
    rec[recfill].refs[k] = -1;
  rec[recfill].band = band;
  rec[recfill].lowmag = 1000.0;
  rec[recfill].himag = -1000.0;  // no magnitude range limitation
  last_recipe[star] = recfill;
  recfill++;
}

static int star_by_label(char *lbl) { // label ends in space, \0 or ')'

  char lotoken[128]; // ad hoc maximum label size
  int li=0;

  while(isspace(*lbl)) lbl++;
  while((!isspace(*lbl)) && ((*lbl)!=')')) {
    lotoken[li++] = *lbl++; // ouch, check lotoken overflow or ugly bug here
  }
  lotoken[li] = 0;

  for (li=0; li<starfill; li++)
    if (!strcmp(lotoken,label[li]))
      return li;

  return -1; // no such star
}

static int _read_shortcat(char *fn) { // 0 if successful; otherwise setup is unusable
                                     // a wrapper (read_shortcat) will deal with this

  int i,j;
  float t;

  if (initialize()) {
    fprintf(stderr, "Malloc failed.\n");
    exit(1);            // it is likely this can't be dealt with later
                        // better suicide than introduce a source of silent
                        // but persistent malfunction
  }
  if (strlen(fn)>=FN_MAX)
    return 1;
  strcpy(filename, fn);
  fi = fopen(fn,"r");
  if (fi==NULL)
    return 1;

  bufill = 0;
  starfill = -1;
  j = -1; // current star to be filled with data

  

  while (read_token(fi)) {
    if ((!strncmp(token,"VARSTAR",7)) || (!strncmp(token,"CONSTAR",7))) {

      if (starfill>=500)   /*********** U G L Y *****************/
	break;             /*********** H A C K *****************/

      j++; starfill ++;
      stype[j] = VARSTAR;
      ra[j] = 0.0;
      dec[j] = 0.0;
      label[j] = vartype[j] = spectype[j] = varid[j] = gsclink[j] = chart[j] = NULL;
      first_recipe[j] = last_recipe[j] = -1;
      for(i=0; i<N_BANDS; i++)
	mag[i][j] = magmax[i][j] = -1000.0; // no magnitude specified
      
      if (!read_token(fi))
	return 1; // failed to read star name
      label[j] = stringstore(token);

      if (!read_token(fi)) 
	return 1; // failed to read star RA
      ra[j] = atof(token);

      if (!read_token(fi))
	return 1; // failed to read star DEC
      dec[j] = atof(token);

      postol[j] = -1000.0; // unspecified
      continue;
    }

    if(j==-1)
      return 1; // first token must be either VARSTAR or CONSTAR

    if (!strncmp(token,"VARTYPE=",8)) {
      vartype[j]=stringstore(token+8);
      continue;
    }

    if (!strncmp(token,"VARID=",6)) {
      varid[j]=stringstore(token+6);
      continue;
    }

    if (!strncmp(token,"GSCL=",5)) {
      gsclink[j] = stringstore(token+5);
      continue;
    }

    if (!strncmp(token,"GSC2L=",6)) {
      gsc2link[j] = stringstore(token+6);
      continue;
    }

    if (!strncmp(token,"SPECTYPE=",9)) {
      spectype[j] = stringstore(token+9);
      continue;
    }

    if (!strncmp(token,"CHART=",6)) {
      chart[j] = stringstore(token+6);
      continue;
    }

    if (!strncmp(token,"TOLERANCE=",10)) {
      postol[j] = atof(token+10);
      continue;
    }

    if (!strncmp(token,"USE-B-REF=",10)) {  // generic B, our B
      add_unparsed_recipe(j, B_BAND, "USE-B-REF=");
      continue;
    }

    if (!strncmp(token,"USE-V-REF=",10)) { // generic V, aavso V, our V
      add_unparsed_recipe(j, V_BAND, "USE-V-REF=");
      continue; // all this because we can't call a function which rasises an exception!
    }

    if (!strncmp(token,"USE-R-REF=",10)) { // GSC2 N
      add_unparsed_recipe(j, R_BAND, "USE-R-REF=");
      continue;
    }

    if (!strncmp(token,"USE-IR-REF=",11)) { // generic IR
      add_unparsed_recipe(j, IR_BAND, "USE-IR-REF=");
      continue;
    }

    if (!strncmp(token,"USE-P-REF=",10)) { // GSC1 P
      add_unparsed_recipe(j, P_BAND, "USE-P-REF=");
      continue;
    }

    if (!strncmp(token,"USE-VT-REF=",11)) { // Tycho-2 V
      add_unparsed_recipe(j, VT_BAND, "USE-VT-REF=");
      continue;
    }

    if (!strncmp(token,"USE-BT-REF=",11)) { // Tycho-2 B
      add_unparsed_recipe(j, BT_BAND, "USE-BT-REF=");
      continue;
    }

    if (!strncmp(token,"USE-JG-REF=",11)) { // GSC2 J
      add_unparsed_recipe(j, JG_BAND, "USE-JG-REF=");
      continue;
    }

    if (!strncmp(token,"USE-FG-REF=",11)) { // GSC2 F
      add_unparsed_recipe(j, FG_BAND, "USE-FG-REF=");
      continue;
    }

    if (!strncmp(token,"USE-VG-REF=",11)) { // GSC2 V
      add_unparsed_recipe(j, VG_BAND, "USE-VG-REF=");
      continue;
    }

    if (!strncmp(token,"USE-NG-REF=",11)) { // GSC2 N
      add_unparsed_recipe(j, NG_BAND, "USE-NG-REF=");
      continue;
    }

    if (!strncmp(token,"BMAG=",5)) { 
      if (token[5]=='(')
	sscanf(token+5,"(%f %f)",mag[B_BAND]+j,magmax[B_BAND]+j);
      else
	mag[B_BAND][j] = atof(token+5);
      continue;
    }

    if (!strncmp(token,"PMAG=",5)) {
      if (token[5]=='(')
	sscanf(token+5,"(%f %f)",mag[P_BAND]+j,magmax[P_BAND]+j);
      else
	mag[P_BAND][j] = atof(token+5);
      continue;
    }
    
    if (!strncmp(token,"VMAG=",5)) { // means aavso v mag
      if (token[5]=='(')
	sscanf(token+5,"(%f %f)",mag[V_BAND]+j,magmax[V_BAND]+j);
      else {
	mag[V_BAND][j] = atof(token+5);
      }
      continue;
    }
    
    if (!strncmp(token,"RMAG=",5)) {
      if (token[5]=='(')
	sscanf(token+5,"(%f %f)",mag[U_BAND]+j,magmax[R_BAND]+j);
      else
	mag[R_BAND][j] = atof(token+5);
      continue;
    }
    
    if (!strncmp(token,"VTMAG=",6)) {
      if (token[6]=='(')
	sscanf(token+6,"(%f %f)",mag[VT_BAND]+j,magmax[VT_BAND]+j);
      else
	mag[VT_BAND][j] = atof(token+6);
      continue;
    }
    
    if (!strncmp(token,"BTMAG=",6)) {
      if (token[6]=='(')
	sscanf(token+6,"(%f %f)",mag[BT_BAND]+j,magmax[BT_BAND]+j);
      else
	mag[BT_BAND][j] = atof(token+6);
      continue;
    }
    
    if (!strncmp(token,"IRMAG=",6)) {
      if (token[6]=='(')
	sscanf(token+6,"(%f %f)",mag[IR_BAND]+j,magmax[IR_BAND]+j);
      else
	mag[IR_BAND][j] = atof(token+6);
      continue;
    }
    
    if (!strncmp(token,"VGMAG=",6)) {
      if (token[6]=='(')
	sscanf(token+6,"(%f %f)",mag[VG_BAND]+j,magmax[VG_BAND]+j);
      else
	mag[VG_BAND][j] = atof(token+6);
      continue;
    }
    
    if (!strncmp(token,"JGMAG=",6)) {
      if (token[6]=='(')
	sscanf(token+6,"(%f %f)",mag[JG_BAND]+j,magmax[JG_BAND]+j);
      else
	mag[JG_BAND][j] = atof(token+6);
      continue;
    }
    
    if (!strncmp(token,"FGMAG=",6)) {
      if (token[6]=='(')
	sscanf(token+6,"(%f %f)",mag[FG_BAND]+j,magmax[FG_BAND]+j);
      else
	mag[FG_BAND][j] = atof(token+6);
      continue;
    }
    
    if (!strncmp(token,"NGMAG=",6)) {
      if (token[6]=='(')
	sscanf(token+6,"(%f %f)",mag[NG_BAND]+j,magmax[NG_BAND]+j);
      else
	mag[NG_BAND][j] = atof(token+6);
      continue;
    }
    
    
  }

  fclose(fi);

  // identify the labels of all reference stars mentioned in recipes

  for (i=0; i<recfill; i++) {
    if (rec[i].src[0]!='(') { // it is a single reference star
      rec[i].refs[0] = star_by_label(rec[i].src);
      rec[i].nrefs = 1;
      continue;
    }
    // now, it is a list; we assume labels of stars do not start wih '-' or a digit
    j = 1;
    while(isspace(rec[i].src[j])) j++;
    if(isdigit(rec[i].src[j]) || rec[i].src[j]=='-') { // we have mag limits
      rec[i].lowmag = atof(rec[i].src+j);
      while(!isspace(rec[i].src[j])) j++;
      while(isspace(rec[i].src[j])) j++;
      if (!(isdigit(rec[i].src[j]) || rec[i].src[j]=='-'))
	return 1; // error: must always provide two magnitudes (in any order)
      rec[i].himag = atof(rec[i].src+j);
      while(!isspace(rec[i].src[j])) j++;
      if (rec[i].lowmag < rec[i].himag) {
	t = rec[i].lowmag;
	rec[i].lowmag = rec[i].himag;
	rec[i].himag = t;
      }
      while(rec[i].src[j]!=')') {
	while(isspace(rec[i].src[j])) j++;
	if(rec[i].src[j]==')')
	  break;
	// also, a ')' is taken to always end a label 
	rec[i].refs[rec[i].nrefs++] = star_by_label(rec[i].src+j);
	while((!isspace(rec[i].src[j])) && (rec[i].src[j]!=')'))
	  j++;
      }
    }
  }
  return 0;
}

static int read_shortcat(char *fn) { // providing for lack of exception support

  if (!strcmp(fn,filename))
    return 0; // already there

  if (_read_shortcat(fn)) {
    starfill = 0;
    bufill = 0;
    filename[0] = 0;
    if (fi!=NULL)
      fclose(fi);
    return 1; // error
  }

  return 0;
}


// The following functions are the external interface to shorcat
// There names must start with some tag, of course, to avoid collisions...
// See libnut.h for a description of how to call these

int rcipe_n_stars(char *catname) { // -1 on error
  
  if(read_shortcat(catname))
    return -1;
  return starfill;
}

// the user is expected to use the above dimension to allocate
// data for vectors specified below; however, the size of
// the allocated vectors must be provided as a last argument as well

int rcipe_coordinates(char *catname, float *out_ra, float *out_dec, int size) {
  // -1 on error or number of results provided

  int i;

  if(read_shortcat(catname)) // not actually read again if just read
    return -1;
  for(i=0; i<size && i<starfill; i++) {
    out_ra[i] = ra[i];
    out_dec[i] = dec[i];
  }
  return i;
}

int rcipe_magnitudes(char *catname, int band, float *out_m, int size) { 
  // -1 on error or number of results provided

  int i;
  if(read_shortcat(catname))
    return -1; //error
  for(i=0; i<size && i<starfill; i++) {
    out_m[i] = mag[band][i];  // this will be the lowest magnitude of the star
  }
  return i;
}

int rcipe_refs(char *catname, int star, float mg, int band, int *out_r, int size) {

  int i,j;

  if(read_shortcat(catname))
    return -1; //error
  for(i=first_recipe[star]; i<=last_recipe[star]; i++) {
    if ((mg>rec[i].lowmag) || (mg<rec[i].himag))
      continue;
    if (band != rec[i].band)
      continue;
    for(j=0; (j<rec[i].nrefs) && (j<size); j++)
      out_r[j] = rec[i].refs[j];
    return j;
  }
  return 0;
}


int rcipe_varstars(char *catname, int *rstars, int size) {

  int i,j;

  if(read_shortcat(catname))
    return -1; //error

  for(i=0, j=0; (j<size) && (i<starfill); i++)
    if(stype[i]==VARSTAR)
      rstars[j++] = i;
  return j;
}


int rcipe_constars(char *catname, int *rstars, int size)  {

  int i,j;

  if(read_shortcat(catname))
    return -1; //error

  for(i=0, j=0; (j<size) && (i<starfill); i++)
    if(stype[i]==CONSTAR)
      rstars[j++] = i;
  return j;
}


float rcipe_highest_mag(char *catname, int band, int star) {

  if(read_shortcat(catname))
    return -1001.0; // this means ``some error, any error'' 
  if(star>=starfill)
    return -1000.0; // this means ``data absent''
  return magmax[band][star];
}

float rcipe_ra(char *catname, int star)  {

  if(read_shortcat(catname))
    return -1001.0; // this means ``some error, any error'' 
  if(star>=starfill)
    return -1000.0; // this means ``data absent''
  return ra[star];
}

float rcipe_dec(char *catname, int star)  {

  if(read_shortcat(catname))
    return -1001.0; // this means ``some error, any error'' 
  if(star>=starfill)
    return -1000.0; // this means ``data absent''
  return dec[star];
}

float rcipe_mag(char *catname, int star, int band)  {

  if(read_shortcat(catname))
    return -1001.0; // this means ``some error, any error'' 
  if(star>=starfill)
    return -1000.0; // this means ``data absent''
  return mag[band][star];
}


float rcipe_postol(char *catname, int star)  {

  if(read_shortcat(catname))
    return -1001.0; // this means ``some error, any error'' 
  if(star>=starfill)
    return -1000.0; // this means ``data absent''
  return postol[star];
}

int rcipe_type(char *catname, int star)  {

  if(read_shortcat(catname))
    return -1; // this means ``some error, any error'' 
  if(star>=starfill)
    return -2; // this means ``data absent''
  return stype[star];
}


int rcipe_label(char *catname, int star, char *obuf, int size) {

  if(read_shortcat(catname))
    return -1;
  if(star>=starfill)
    return -2;
  if (label[star]==NULL) {
    obuf[0]=0;
    return 0;
  } else
    strncpy(obuf, label[star], size);
  label[size-1] = 0; // just in case
  if (strlen(label[star])>=size)
    return -3; // truncated
  return 0;
}


int rcipe_chart(char *catname, int star, char *obuf, int size) {

  if(read_shortcat(catname))
    return -1;
  if(star>=starfill)
    return -2;
  if (chart[star]==NULL) {
    obuf[0]=0;
    return 0;
  } else
    strncpy(obuf, chart[star], size);
  label[size-1] = 0; // just in case
  if (strlen(chart[star])>=size)
    return -3; // truncated
  return 0;
}

int rcipe_gsclink(char *catname, int star, char *obuf, int size) {

  if(read_shortcat(catname))
    return -1;
  if(star>=starfill)
    return -2;
  if (gsclink[star]==NULL) {
    obuf[0]=0;
    return 0;
  } else
    strncpy(obuf, gsclink[star], size);
  gsclink[size-1] = 0; // just in case
  if (strlen(gsclink[star])>=size)
    return -3; // truncated
  return 0;
}


int rcipe_gsc2link(char *catname, int star, char *obuf, int size) {

  if(read_shortcat(catname))
    return -1;
  if(star>=starfill)
    return -2;
  if (gsc2link[star]==NULL) {
    obuf[0]=0;
    return 0;
  } else
    strncpy(obuf, gsc2link[star], size);
  gsc2link[size-1] = 0; // just in case
  if (strlen(gsc2link[star])>=size)
    return -3; // truncated
  return 0;
}

int rcipe_spectype(char *catname, int star, char *obuf, int size) {

  if(read_shortcat(catname))
    return -1;
  if(star>=starfill)
    return -2;
  if (spectype[star]==NULL) {
    obuf[0]=0;
    return 0;
  } else
    strncpy(obuf, spectype[star], size);
  spectype[size-1] = 0; // just in case
  if (strlen(spectype[star])>=size)
    return -3; // truncated
  return 0;
}


int rcipe_varid(char *catname, int star, char *obuf, int size) {

  if(read_shortcat(catname))
    return -1;
  if(star>=starfill)
    return -2;
  if (varid[star]==NULL) {
    obuf[0]=0;
    return 0;
  } else
    strncpy(obuf, varid[star], size);
  label[size-1] = 0; // just in case
  if (strlen(varid[star])>=size)
    return -3; // truncated
  return 0;
}

int rcipe_vartype(char *catname, int star, char *obuf, int size)  {

  if(read_shortcat(catname))
    return -1;
  if(star>=starfill)
    return -2;
  if (vartype[star]==NULL) {
    obuf[0]=0;
    return 0;
  } else
    strncpy(obuf, vartype[star], size);
  label[size-1] = 0; // just in case
  if (strlen(vartype[star])>=size)
    return -3; // truncated
  return 0;
}

int rcipe_named(char *catname, char *lbl) {

  int i;

  if(read_shortcat(catname))
    return -1;
  for(i=0; i<starfill; i++)
    if(!strcmp(lbl, label[i]))
      return i;
  return -2; // not found
}

/*  The following function is lifted and modified from worldpos.c
    which had the copyright notice which follows

    The xypic function was changed to perform only a TAN projection
    of unrotated coordinates.

    worldpos.c -- WCS Algorithms from Classic AIPS.
    Copyright (C) 1994
    Associated Universities, Inc. Washington DC, USA.
   
    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Library General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.
   
    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
    License for more details.
   
    You should have received a copy of the GNU Library General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
   
    Correspondence concerning AIPS should be addressed as follows:
           Internet email: aipsmail@nrao.edu
           Postal address: AIPS Group
                           National Radio Astronomy Observatory
                           520 Edgemont Road
                           Charlottesville, VA 22903-2475 USA

                 -=-=-=-=-=-=-

    These two ANSI C functions, worldpos() and xypix(), perform
    forward and reverse WCS computations for 8 types of projective
    geometries ("-SIN", "-TAN", "-ARC", "-NCP", "-GLS", "-MER", "-AIT"
    and "-STG"):

        worldpos() converts from pixel location to RA,Dec 
        xypix()    converts from RA,Dec         to pixel location   

    where "(RA,Dec)" are more generically (long,lat). These functions
    are based on the WCS implementation of Classic AIPS, an
    implementation which has been in production use for more than ten
    years. See the two memos by Eric Greisen

        ftp://fits.cv.nrao.edu/fits/documents/wcs/aips27.ps.Z
	ftp://fits.cv.nrao.edu/fits/documents/wcs/aips46.ps.Z

    for descriptions of the 8 projective geometries and the
    algorithms.  Footnotes in these two documents describe the
    differences between these algorithms and the 1993-94 WCS draft
    proposal (see URL below). In particular, these algorithms support
    ordinary field rotation, but not skew geometries (CD or PC matrix
    cases). Also, the MER and AIT algorithms work correctly only for
    CRVALi=(0,0). Users should note that GLS projections with yref!=0
    will behave differently in this code than in the draft WCS
    proposal.  The NCP projection is now obsolete (it is a special
    case of SIN).  WCS syntax and semantics for various advanced
    features is discussed in the draft WCS proposal by Greisen and
    Calabretta at:
    
        ftp://fits.cv.nrao.edu/fits/documents/wcs/wcs.all.ps.Z
    
                -=-=-=-

    The original version of this code was Emailed to D.Wells on
    Friday, 23 September by Bill Cotton <bcotton@gorilla.cv.nrao.edu>,
    who described it as a "..more or less.. exact translation from the
    AIPSish..". Changes were made by Don Wells <dwells@nrao.edu>
    during the period October 11-13, 1994:
    1) added GNU license and header comments
    2) added testpos.c program to perform extensive circularity tests
    3) changed float-->double to get more than 7 significant figures
    4) testpos.c circularity test failed on MER and AIT. B.Cotton
       found that "..there were a couple of lines of code [in] the wrong
       place as a result of merging several Fortran routines." 
    5) testpos.c found 0h wraparound in xypix() and worldpos().
    6) E.Greisen recommended removal of various redundant if-statements,
       and addition of a 360d difference test to MER case of worldpos(). 
*/

int tan_xypix(xpos, ypos, xref, yref, xrefpix, yrefpix, xinc, yinc, rot,
		     xpix, ypix)
double xpos;
double ypos;
double xref;
double yref; 
double xrefpix;
double yrefpix;
double xinc;
double yinc;
double rot;
double *xpix;
double *ypix;
/*-----------------------------------------------------------------------*/
/* routine to determine accurate pixel coordinates for an RA and Dec     */
/* returns 0 if successful otherwise:                                    */
/* 1 = angle too large for projection;                                   */
/* 2 = bad values                                                        */
/* does: -SIN, -TAN, -ARC, -NCP, -GLS, -MER, -AIT projections            */
/* anything else is linear                                               */
/* Input:                                                                */
/*   d   xpos    x (RA) coordinate (deg)                                 */
/*   d   ypos    y (dec) coordinate (deg)                                */
/*   d   xref    x reference coordinate value (deg)                      */
/*   d   yref    y reference coordinate value (deg)                      */
/*   f   xrefpix x reference pixel                                       */
/*   f   yrefpix y reference pixel                                       */
/*   f   xinc    x coordinate increment (deg)                            */
/*   f   yinc    y coordinate increment (deg)                            */
/*   f   rot     rotation (deg)  (from N through E)                      */
/*   c  *type    projection type code e.g. "-SIN";                       */
/* Output:                                                               */
/*   f  *xpix    x pixel number  (RA or long without rotation)           */
/*   f  *ypiy    y pixel number  (dec or lat without rotation)           */
/*-----------------------------------------------------------------------*/
{
  double dx, dy, dz, r, ra0, dec0, ra, dec, coss, sins, dt, da, dd, sint;
  double l, m=0, geo1, geo2, geo3, sinr, cosr;
  double cond2r=1.745329252e-2, deps=1.0e-5, twopi=6.28318530717959;
  int   i, itype;
  //  static char ctypes[8][5] ={"-SIN","-TAN","-ARC","-NCP", "-GLS", "-MER",
  //      "-AIT", "-STG"};

  /* 0h wrap-around tests added by D.Wells 10/12/94: */
  dt = (xpos - xref);
  if (dt >  180) xpos -= 360;
  if (dt < -180) xpos += 360;
  /* NOTE: changing input argument xpos is OK (call-by-value in C!) */

/* default values - linear */
  dx = xpos - xref;
  dy = ypos - yref;
  dz = 0.0;
/*  Correct for rotation */
  r = rot * cond2r;
  cosr = cos (r); 
  sinr = sin (r);
  dz = dx*cosr + dy*sinr;
  dy = dy*cosr - dx*sinr;
  dx = dz;
/*     check axis increments - bail out if either 0 */
  if ((xinc==0.0) || (yinc==0.0)) {*xpix=0.0; *ypix=0.0; return 2;}
/*     convert to pixels  */
  *xpix = dx / xinc + xrefpix;
  *ypix = dy / yinc + yrefpix;

/* Non linear position */
  ra0 = xref * cond2r;
  dec0 = yref * cond2r;
  ra = xpos * cond2r;
  dec = ypos * cond2r;

/* compute direction cosine */
  coss = cos (dec);
  sins = sin (dec);
  l = sin(ra-ra0) * coss;
  sint = sins * sin(dec0) + coss * cos(dec0) * cos(ra-ra0);

/* process by case, but there is a single case: TAN  */

  if (sint<=0.0) return 1;
  m = sins * sin(dec0) + coss * cos(dec0) * cos(ra-ra0);
  l = l / m;
  m = (sins * cos(dec0) - coss * sin(dec0) * cos(ra-ra0)) / m;


/*   back to degrees  */
  dx = l / cond2r;
  dy = m / cond2r;
/*  Correct for rotation */
  dz = dx*cosr + dy*sinr;
  dy = dy*cosr - dx*sinr;
  dx = dz;
/*     convert to pixels  */
  *xpix = dx / xinc + xrefpix;
  *ypix = dy / yinc + yrefpix;
  return 0;
}  

/* end of section lifted from worldpos.c */


int rcipe_xy(char *catname, float *out_x, float *out_y,
		     float deg_per_pix, int size) {
  int nn,i;
  double ocx,ocy;

  nn = rcipe_coordinates(catname, out_x, out_y, size);
  if (nn<1)
    return nn;

  // the first object, which always exists from here on, and
  // which is commonly the target object in the center of the image
  // and also to be found in the center of the shorcat
  // will be taken as the center at pixel position 0,0

  for(i=1; i<nn; i++) {
    tan_xypix((double)(out_x[i]),(double)(out_y[i]),(double)(out_x[0]),(double)(out_y[0]),
	      (double)0.0, (double)0.0,
              (double)deg_per_pix, (double)deg_per_pix,  (double)0.0, 
	      &ocx, &ocy);
    out_x[i] = ocx;
    out_y[i] = ocy;
  }

  out_x[0] = 0.0;
  out_y[0] = 0.0;

  return nn;
}


void rcipe_test(char *catname) {

  float raa[1000];
  float deca[1000];
  float mag[1000];
  int rf[1000];
  int i,r;
  char bf[100];

  printf("\n\n*** Testing shorcat library for %s\n\n", catname);

  r = rcipe_coordinates(catname, raa, deca, 1000);

  printf("    Number of objects = %d\n", r);
  for (i=0; i<10; i++)
    printf("       coords of %d: %f %f\n", i, raa[i], deca[i]);
  rcipe_magnitudes(catname, V_BAND, mag, 1000);
  printf("\n    Mag of 5 in V=%f\n", mag[5]);
  printf("    Mag of 0 in V=%f\n", mag[0]);
  printf("    Hi mag of 0 in B=%f\n", rcipe_highest_mag(catname, B_BAND, 0));
  r = rcipe_refs(catname, 0, 15.0, V_BAND, rf, 1000);
  printf("\n    Number of reference stars for star 0 = %d\n", r);
  for (i=0; i<r; i++)
    printf("       %d. is star %d\n", i, rf[i]);
  printf("\n    Components of star 0\n");
  if(rcipe_label(catname,0,bf,1000)>=0)
    printf("      Label: %s\n", bf);
  else
    printf("      Label: failed to obtain\n");
  if(rcipe_spectype(catname,0,bf,1000)>=0)
    printf("      Spectype: %s\n", bf);
  else
    printf("      Spectype: failed to obtain\n");
  if(rcipe_vartype(catname,0,bf,1000)>=0)
    printf("      Vartype: %s\n", bf);
  else
    printf("      Vartype: failed to obtain\n");
  if(rcipe_varid(catname,0,bf,1000)>=0)
    printf("      Varid: %s\n", bf);
  else
    printf("      Varid: failed to obtain\n");
  if(rcipe_gsclink(catname,0,bf,1000)>=0)
    printf("      GSC link: %s\n", bf);
  else
    printf("      GSC link: failed to obtain\n");

  if(rcipe_ra(catname,0) >= -999.0)
    printf("      RA=%f\n", rcipe_ra(catname,0));
  else
    printf("      RA=(not available)\n");
  if(rcipe_dec(catname,0) >= -999.0)
    printf("      DEC=%f\n", rcipe_dec(catname,0));
  else
    printf("      DEC=(not available)\n");
  if(rcipe_mag(catname,0,B_BAND) >= -999.0)
    printf("      MAG (B)=%f\n", rcipe_mag(catname,0,B_BAND));
  else
    printf("      MAG (B)=(not available)\n");
  if(rcipe_highest_mag(catname,0,B_BAND) >= -999.0)
    printf("      HIMAG (B)=%f\n", rcipe_highest_mag(catname,0,B_BAND));
  else
    printf("      HIMAG (B)=(not available)\n");
  
}

/*
main(short argc, char **argv) {
  rcipe_test(argv[1]);
}
*/

